/*
 * Copyright 2015 Luca Beltrame <lbeltrame@kde.org>
 *
 * This file is part of Danbooru Client.
 *
 * Danbooru Client is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Danbooru Client is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Danbooru Client. If not, see <http://www.gnu.org/licenses/>.
 */

#include "konachan.h"
#include "danboorupost.h"
#include "danboorupool.h"
#include "danboorutag.h"
#include "utils.h"
#include "libdanbooru_debug.h"

// KF5

#include <KIO/Job>
#include <KIO/Scheduler>
#include <KImageCache>

using KIO::StoredTransferJob;

namespace Danbooru {

///////
// URIs
///////

const QLatin1String KonachanDanbooruService::postUri() const {
    return QLatin1String("post/index.json");
}

const QLatin1String KonachanDanbooruService::poolUri() const {
    return QLatin1String("pool/index.json");
}

const QLatin1String KonachanDanbooruService::artistUri() const {
    return QLatin1String("artist/index.json");
}

const QLatin1String KonachanDanbooruService::tagUri() const {
    return QLatin1String("tag/index.xml");
}

const QLatin1String KonachanDanbooruService::poolDataUri() const {
    return QLatin1String("pool/show.xml");
}

const QLatin1String KonachanDanbooruService::relatedTagUri() const {
    return QLatin1String("tag/related.json");
}

void KonachanDanbooruService::getPostList() {
    // We can't fetch more than 100 items, API limitation

    QMap<QString, QString> parameters;

    parameters.insert("limit", QString::number(m_maxPosts));
    parameters.insert("page", QString::number(m_currentPage));

    QUrl danbooruUrl = requestUrl(m_url, postUri(), m_username,
                                  m_password, parameters, m_tags);

    qCDebug(LIBDANBOORU) << "Final constructed post URL" << danbooruUrl;

    KIO::StoredTransferJob *job = KIO::storedGet(danbooruUrl, KIO::NoReload,
                                  KIO::HideProgressInfo);

    // This job can use JSON data
    job->setProperty("needsXML", false);

    connect(job, &KIO::StoredTransferJob::result, this, &KonachanDanbooruService::processPostList);

}

void KonachanDanbooruService::processPostList(KJob *job)
{

    qCDebug(LIBDANBOORU) << "Got post data OK";

    if (job->error()) {
        Q_EMIT(downloadError(job->errorString()));
    }

    StoredTransferJob *jobResult = qobject_cast<StoredTransferJob *>(job);

    if (jobResult == 0) {
        Q_EMIT(downloadError(QString("Internal error")));
        return;

    }

    QByteArray data = jobResult->data();

    bool ok;
    bool needsXML = job->property("needsXML").toBool();

    QList<QVariant> postList;

    if (needsXML) {
        // Special cases for pools
        postList = parseDanbooruResult(data, QString("post"), &ok);
    } else {
        postList = parseDanbooruResult(data, &ok).toList();
    }

    if (!ok) {
        Q_EMIT(downloadError(QString("Unable to decode data")));
        return;
    }

    // How many posts do we have to fetch?

    if (postList.isEmpty()) {
        qCDebug(LIBDANBOORU) << "No posts found";
        Q_EMIT(postDownloadFinished());
        return;
    }

    m_postsToFetch = postList.length();

    for (auto element : postList) {

        QVariantMap map = element.toMap();
        DanbooruPost *post = new DanbooruPost(map);

        // Remove unwanted posts

        if (isPostBlacklisted(post, m_blacklist, m_maxRating)) {
            m_postsToFetch--;
            delete post;
            continue;
        }

        QPixmap pix;

        qCDebug(LIBDANBOORU) << "About to donwload images";

        if (m_cache && m_cache->findPixmap(post->thumbnailUrl().url(), &pix)) {

            post->setPixmap(pix);
            Q_EMIT(postDownloaded(post));
            m_postsToFetch--;

            if (m_postsToFetch == 0) {
                qCDebug(LIBDANBOORU) << "Post download finished";
                Q_EMIT(postDownloadFinished());
            }

        } else {

            StoredTransferJob *pixmapJob = KIO::storedGet(post->thumbnailUrl(),
                                           KIO::NoReload, KIO::HideProgressInfo);

            // We don't want to overload the servers, so set some rational
            // priority

            KIO::Scheduler::setJobPriority(static_cast<KIO::SimpleJob *>(job), 1);

            QVariant variant;
            variant.setValue(post);

//             pixmapJob->setProperty("danbooruPost", variant);

            connect(pixmapJob, &StoredTransferJob::result, [post, this, pix](KJob * job) mutable {

                if (job->error())
                {
                    Q_EMIT(downloadError(job->errorString()));
                    return;
                }

                StoredTransferJob *jobResult = qobject_cast<StoredTransferJob *>(job);

                if (!pix.loadFromData(jobResult->data()))
                {
                    Q_EMIT(downloadError(QString("Pixmap data could not be loaded")));
                    return;
                }

                post->setPixmap(pix);

                if (m_cache)
                {
                    //qCDebug(LIBDANBOORU) << "Inserting item in cache";
                    m_cache->insertPixmap(post->thumbnailUrl().url(), pix);
                }

                m_postsToFetch--; // One less post to do

                qCDebug(LIBDANBOORU) << "Current posts remaining: " << m_postsToFetch;
                Q_EMIT(postDownloaded(post));

                if (m_postsToFetch == 0)
                {
                    qCDebug(LIBDANBOORU) << "Post download finished";
                    Q_EMIT(postDownloadFinished());
                }

            });

        }

    }

}

} //namespace Danbooru


